@@ -33,6 +33,8 @@
 # we use some perl unicode features
 require 5.006;
 
+use bytes;
+
 use vars qw($CODE $NAME $CATEGORY $COMBINING_CLASSES $BIDI_CATEGORY $DECOMPOSITION $DECIMAL_VALUE $DIGIT_VALUE $NUMERIC_VALUE $MIRRORED $OLD_NAME $COMMENT $UPPER $LOWER $TITLE $BREAK_CODE $BREAK_CATEGORY $BREAK_NAME $CASE_CODE $CASE_LOWER $CASE_TITLE $CASE_UPPER $CASE_CONDITION);
 
 
@@ -109,37 +111,43 @@ $FOLDING_MAPPING = 2;
 
 %break_mappings =
     (
-     'BK' => "G_UNICODE_BREAK_MANDATORY",
-     'CR' => "G_UNICODE_BREAK_CARRIAGE_RETURN",
-     'LF' => "G_UNICODE_BREAK_LINE_FEED",
-     'CM' => "G_UNICODE_BREAK_COMBINING_MARK",
-     'SG' => "G_UNICODE_BREAK_SURROGATE",
-     'ZW' => "G_UNICODE_BREAK_ZERO_WIDTH_SPACE",
-     'IN' => "G_UNICODE_BREAK_INSEPARABLE",
-     'GL' => "G_UNICODE_BREAK_NON_BREAKING_GLUE",
-     'CB' => "G_UNICODE_BREAK_CONTINGENT",
-     'SP' => "G_UNICODE_BREAK_SPACE",
+     'AI' => "G_UNICODE_BREAK_AMBIGUOUS",
+     'AL' => "G_UNICODE_BREAK_ALPHABETIC",
+     'B2' => "G_UNICODE_BREAK_BEFORE_AND_AFTER",
      'BA' => "G_UNICODE_BREAK_AFTER",
      'BB' => "G_UNICODE_BREAK_BEFORE",
-     'B2' => "G_UNICODE_BREAK_BEFORE_AND_AFTER",
-     'HY' => "G_UNICODE_BREAK_HYPHEN",
-     'NS' => "G_UNICODE_BREAK_NON_STARTER",
-     'OP' => "G_UNICODE_BREAK_OPEN_PUNCTUATION",
+     'BK' => "G_UNICODE_BREAK_MANDATORY",
+     'CB' => "G_UNICODE_BREAK_CONTINGENT",
      'CL' => "G_UNICODE_BREAK_CLOSE_PUNCTUATION",
-     'QU' => "G_UNICODE_BREAK_QUOTATION",
+     'CM' => "G_UNICODE_BREAK_COMBINING_MARK",
+     'CP' => "G_UNICODE_BREAK_CLOSE_PARANTHESIS",
+     'CR' => "G_UNICODE_BREAK_CARRIAGE_RETURN",
      'EX' => "G_UNICODE_BREAK_EXCLAMATION",
+     'GL' => "G_UNICODE_BREAK_NON_BREAKING_GLUE",
+     'H2' => "G_UNICODE_BREAK_HANGUL_LV_SYLLABLE",
+     'H3' => "G_UNICODE_BREAK_HANGUL_LVT_SYLLABLE",
+     'HY' => "G_UNICODE_BREAK_HYPHEN",
      'ID' => "G_UNICODE_BREAK_IDEOGRAPHIC",
-     'NU' => "G_UNICODE_BREAK_NUMERIC",
+     'IN' => "G_UNICODE_BREAK_INSEPARABLE",
      'IS' => "G_UNICODE_BREAK_INFIX_SEPARATOR",
-     'SY' => "G_UNICODE_BREAK_SYMBOL",
-     'AL' => "G_UNICODE_BREAK_ALPHABETIC",
-     'PR' => "G_UNICODE_BREAK_PREFIX",
-     'PO' => "G_UNICODE_BREAK_POSTFIX",
-     'SA' => "G_UNICODE_BREAK_COMPLEX_CONTEXT",
-     'AI' => "G_UNICODE_BREAK_AMBIGUOUS",
+     'JL' => "G_UNICODE_BREAK_HANGUL_L_JAMO",
+     'JT' => "G_UNICODE_BREAK_HANGUL_T_JAMO",
+     'JV' => "G_UNICODE_BREAK_HANGUL_V_JAMO",
+     'LF' => "G_UNICODE_BREAK_LINE_FEED",
      'NL' => "G_UNICODE_BREAK_NEXT_LINE",
+     'NS' => "G_UNICODE_BREAK_NON_STARTER",
+     'NU' => "G_UNICODE_BREAK_NUMERIC",
+     'OP' => "G_UNICODE_BREAK_OPEN_PUNCTUATION",
+     'PO' => "G_UNICODE_BREAK_POSTFIX",
+     'PR' => "G_UNICODE_BREAK_PREFIX",
+     'QU' => "G_UNICODE_BREAK_QUOTATION",
+     'SA' => "G_UNICODE_BREAK_COMPLEX_CONTEXT",
+     'SG' => "G_UNICODE_BREAK_SURROGATE",
+     'SP' => "G_UNICODE_BREAK_SPACE",
+     'SY' => "G_UNICODE_BREAK_SYMBOL",
      'WJ' => "G_UNICODE_BREAK_WORD_JOINER",
-     'XX' => "G_UNICODE_BREAK_UNKNOWN"
+     'XX' => "G_UNICODE_BREAK_UNKNOWN",
+     'ZW' => "G_UNICODE_BREAK_ZERO_WIDTH_SPACE"
      );
 
 # Title case mappings.
@@ -168,21 +176,20 @@ elsif (@ARGV && $ARGV[0] eq '-both')
 
 if (@ARGV != 2) {
     $0 =~ s@.*/@@;
-    die "\nUsage: $0 [-decomp | -both] UNICODE-VERSION DIRECTORY\n\n       DIRECTORY should contain the following Unicode data files:\n       UnicodeData.txt, LineBreak.txt, SpecialCasing.txt, CaseFolding.txt,\n       CompositionExclusions.txt, BidiMirroring.txt\n\n";
+    die "\nUsage: $0 [-decomp | -both] UNICODE-VERSION DIRECTORY\n\n       DIRECTORY should contain the following Unicode data files:\n       UnicodeData.txt, LineBreak.txt, SpecialCasing.txt, CaseFolding.txt,\n       CompositionExclusions.txt\n\n";
 }
 
-my ($unicodedatatxt, $linebreaktxt, $specialcasingtxt, $casefoldingtxt, $compositionexclusionstxt, $bidimirroringtxt);
+my ($unicodedatatxt, $linebreaktxt, $specialcasingtxt, $casefoldingtxt, $compositionexclusionstxt);
 
 my $d = $ARGV[1];
 opendir (my $dir, $d) or die "Cannot open Unicode data dir $d: $!\n";
 for my $f (readdir ($dir))
 {
-    $unicodedatatxt = "$d/$f" if ($f =~ /UnicodeData.*\.txt/);
-    $linebreaktxt = "$d/$f" if ($f =~ /LineBreak.*\.txt/);
-    $specialcasingtxt = "$d/$f" if ($f =~ /SpecialCasing.*\.txt/);
-    $casefoldingtxt = "$d/$f" if ($f =~ /CaseFolding.*\.txt/);
-    $compositionexclusionstxt = "$d/$f" if ($f =~ /CompositionExclusions.*\.txt/);
-    $bidimirroringtxt = "$d/$f" if ($f =~ /BidiMirroring.*\.txt/);
+    $unicodedatatxt = "$d/$f" if ($f =~ /^UnicodeData.*\.txt/);
+    $linebreaktxt = "$d/$f" if ($f =~ /^LineBreak.*\.txt/);
+    $specialcasingtxt = "$d/$f" if ($f =~ /^SpecialCasing.*\.txt/);
+    $casefoldingtxt = "$d/$f" if ($f =~ /^CaseFolding.*\.txt/);
+    $compositionexclusionstxt = "$d/$f" if ($f =~ /^CompositionExclusions.*\.txt/);
 }
 
 defined $unicodedatatxt or die "Did not find UnicodeData file";
@@ -190,7 +197,6 @@ defined $linebreaktxt or die "Did not find LineBreak file";
 defined $specialcasingtxt or die "Did not find SpecialCasing file";
 defined $casefoldingtxt or die "Did not find CaseFolding file";
 defined $compositionexclusionstxt or die "Did not find CompositionExclusions file";
-defined $bidimirroringtxt or die "Did not find BidiMirroring file";
 
 print "Creating decomp table\n" if ($do_decomp);
 print "Creating property table\n" if ($do_props);
@@ -291,6 +297,7 @@ while (<INPUT>)
     chop;
 
     next if /^#/;
+    next if /^$/;
 
     s/\s*#.*//;
     
@@ -481,23 +488,6 @@ while (<INPUT>)
 
 close INPUT;
 
-open (INPUT, "< $bidimirroringtxt") || exit 1;
-
-my @bidimirror;
-while (<INPUT>)
-{
-    chomp;
-
-    next if /^#/;
-    next if /^\s*$/;
-
-    s/\s*#.*//;
-
-    @fields = split ('\s*;\s*', $_, 30);
-
-    push @bidimirror, [hex ($fields[0]), hex ($fields[1])];
-}
- 
 if ($do_props) {
     &print_tables ($last_code)
 }
@@ -516,7 +506,6 @@ sub length_in_bytes
 {
     my ($string) = @_;
 
-    use bytes;
     return length $string;
 }
 
@@ -674,21 +663,6 @@ sub print_tables
     &output_special_case_table (\*OUT);
     &output_casefold_table (\*OUT);
 
-    print OUT "static const struct {\n";
-    print OUT "    gunichar ch;\n";
-    print OUT "    gunichar mirrored_ch;\n";
-    print OUT "} bidi_mirroring_table[] =\n";
-    print OUT "{\n";
-    $first = 1;
-    foreach $item (@bidimirror)
-    {
-        print OUT ",\n" unless $first;
-        $first = 0;
-        printf OUT "  { 0x%04x, 0x%04x }", $item->[0], $item->[1];
-        $bytes_out += 8;
-    }
-    print OUT "\n};\n\n";
-
     print OUT "#endif /* CHARTABLES_H */\n";
 
     close (OUT);
@@ -918,6 +892,9 @@ sub print_line_break
     print OUT "#ifndef BREAKTABLES_H\n";
     print OUT "#define BREAKTABLES_H\n\n";
 
+    print OUT "#include <glib/gtypes.h>\n";
+    print OUT "#include <glib/gunicode.h>\n\n";
+
     print OUT "#define G_UNICODE_DATA_VERSION \"$ARGV[0]\"\n\n";
 
     printf OUT "#define G_UNICODE_LAST_CHAR 0x%04X\n\n", $last;
@@ -1024,8 +1001,8 @@ sub add_special_case
 	       (map { hex ($_) } split /\s+/, $field1),
                0,
                (map { hex ($_) } split /\s+/, $field2));
-    $result = "";
 
+    $result = "";
 
     for $value (@values) {
 	$result .= pack ("U", $value);  # to utf-8
@@ -1229,12 +1206,9 @@ sub output_composition_table
 
     # Output first singletons
 
-    print OUT "static const guint16 compose_first_single[][2] = {\n";
+    print OUT "static const gunichar compose_first_single[][2] = {\n";
     $i = 0;				     
     for $record (@first_singletons) {
-        if ($record->[1] > 0xFFFF or $record->[2] > 0xFFFF) {
-            die "time to switch compose_first_single to gunichar" ;
-        }
 	print OUT ",\n" if $i++ > 0;
 	printf OUT " { %#06x, %#06x }", $record->[1], $record->[2];
     }
